@@ -66,8 +66,7 @@ class SFConfig(BaseSettings):
         (
             '/srv/shakenfist/venv/bin/gunicorn --workers %(workers)d --bind 0.0.0.0:%(port)d '
             '--log-syslog --log-syslog-prefix sf --timeout %(timeout)s --name "%(name)s" '
-            '--pid /var/run/sf/gunicorn.pid --worker-class gevent '
-            'shakenfist.external_api.app:app'
+            '--pid /var/run/sf/gunicorn.pid shakenfist.external_api.app:app'
         ),
         description='The gunicorn command line to use'
     )
@@ -271,6 +270,10 @@ class SFConfig(BaseSettings):
         False, description='Add method name and module line number to log messages'
     )
 
+    EXCESSIVE_ETCD_CACHE_LOGGING: bool = Field(
+        False, description='Record detailed information about etcd cache performance.'
+    )
+
     class Config:
         env_prefix = 'SHAKENFIST_'
 
