@@ -10,8 +10,9 @@ from shakenfist import baseobject
 from shakenfist.config import config
 from shakenfist.daemons import daemon
 from shakenfist import db
+from shakenfist import etcd
 from shakenfist import logutil
-from shakenfist import virt
+from shakenfist import instance
 
 
 LOG, _ = logutil.setup(__name__)
@@ -41,8 +42,22 @@ def observe(path, instance_uuid):
     # noisy after the login prompt.)
     os.lseek(fd, max(0, os.fstat(fd).st_size - 4096), os.SEEK_SET)
 
+    # Record how long the file is, because we need to detect truncations and
+    # re-open.
+    previous_size = os.stat(path).st_size
+
     buffer = ''
     while True:
+        # Detect file truncations, and die if we see one. We will be restarted
+        # by the monitor process.
+        if not os.path.exists(path):
+            return
+        size = os.stat(path).st_size
+        if size < previous_size:
+            return
+        previous_size = size
+
+        # Read data, os.read() is non-blocking by the way.
         d = os.read(fd, 1024).decode('utf-8', errors='ignore')
         if d:
             buffer += d
@@ -60,15 +75,15 @@ def observe(path, instance_uuid):
                                          None, None, trigger)
         else:
             # Only pause if there was no data to read
-            time.sleep(1)
+            time.sleep(0.2)
 
 
 class Monitor(daemon.Daemon):
     def run(self):
-        LOG.info('Starting Monitor Daemon')
+        LOG.info('Starting')
         observers = {}
 
-        while True:
+        while not self.exit.is_set():
             # Cleanup terminated observers
             all_observers = list(observers.keys())
             for instance_uuid in all_observers:
@@ -81,30 +96,35 @@ class Monitor(daemon.Daemon):
                         'instance', instance_uuid, 'trigger monitor', 'crashed', None, None)
                     del observers[instance_uuid]
 
-            # Start missing observers
+            # Audit desired observers
             extra_instances = list(observers.keys())
+            missing_instances = []
 
-            for inst in virt.Instances([virt.this_node_filter,
-                                        partial(baseobject.state_filter, ['created'])]):
-                if inst.uuid in extra_instances:
-                    extra_instances.remove(inst.uuid)
+            with etcd.ThreadLocalReadOnlyCache():
+                for inst in instance.Instances([
+                        instance.this_node_filter,
+                        partial(baseobject.state_filter, [instance.Instance.STATE_CREATED])]):
+                    if inst.uuid in extra_instances:
+                        extra_instances.remove(inst.uuid)
 
-                if inst.uuid not in observers:
-                    console_path = os.path.join(config.get('STORAGE_PATH'),
-                                                'instances',
-                                                inst.uuid,
-                                                'console.log')
-                    p = multiprocessing.Process(
-                        target=observe, args=(console_path, inst.uuid),
-                        name='%s-%s' % (daemon.process_name('triggers'),
-                                        inst.uuid))
-                    p.start()
+                    if inst.uuid not in observers:
+                        missing_instances.append(inst.uuid)
 
-                    observers[inst.uuid] = p
-                    LOG.with_instance(inst.uuid).info(
-                        'Started trigger observer')
-                    db.add_event(
-                        'instance', inst.uuid, 'trigger monitor', 'started', None, None)
+            # Start missing observers
+            for instance_uuid in missing_instances:
+                console_path = os.path.join(
+                    config.STORAGE_PATH, 'instances', instance_uuid, 'console.log')
+                p = multiprocessing.Process(
+                    target=observe, args=(console_path, instance_uuid),
+                    name='%s-%s' % (daemon.process_name('triggers'),
+                                    instance_uuid))
+                p.start()
+
+                observers[instance_uuid] = p
+                LOG.with_instance(instance_uuid).info(
+                    'Started trigger observer')
+                db.add_event(
+                    'instance', instance_uuid, 'trigger monitor', 'started', None, None)
 
             # Cleanup extra observers
             for instance_uuid in extra_instances:
@@ -121,4 +141,8 @@ class Monitor(daemon.Daemon):
                 db.add_event(
                     'instance', instance_uuid, 'trigger monitor', 'finished', None, None)
 
-            time.sleep(1)
+            self.exit.wait(1)
+
+        # No longer running, clean up all trigger deaemons
+        for instance_uuid in observers:
+            os.kill(observers[instance_uuid].pid, signal.SIGKILL)
