@@ -1,22 +1,21 @@
 # Make scheduling decisions
 
-import copy
-from functools import partial
+from collections import defaultdict
+import math
 import random
 import time
 import uuid
 
-from shakenfist import baseobject
 from shakenfist.config import config
-from shakenfist import db
+from shakenfist.constants import GiB
+from shakenfist import etcd
 from shakenfist import exceptions
-from shakenfist import images
+from shakenfist import instance
 from shakenfist import logutil
 from shakenfist import networkinterface
 from shakenfist.node import (
     Nodes, active_states_filter as node_active_states_filter)
-from shakenfist import util
-from shakenfist import virt
+from shakenfist.util import general as util_general
 
 
 LOG, _ = logutil.setup(__name__)
@@ -40,7 +39,7 @@ def get_network_node():
             CACHED_NETWORK_NODE = n
             return CACHED_NETWORK_NODE
 
-    return None
+    raise exceptions.NoNetworkNode('Cannot find network node')
 
 
 class Scheduler(object):
@@ -60,16 +59,19 @@ class Scheduler(object):
         metrics = {}
 
         for n in Nodes([node_active_states_filter]):
-            node_name = n.uuid
             try:
-                new_metrics = db.get_metrics(node_name)
-                self.log.with_object(n).debug(
-                    'Metrics for node: %s', new_metrics)
+                new_metrics = etcd.get('metrics', n.uuid, None)
                 if new_metrics:
-                    metrics[node_name] = new_metrics
+                    new_metrics = new_metrics.get('metrics', {})
                 else:
                     self.log.with_object(n).warning(
                         'Empty metrics from database for node')
+                    new_metrics = {}
+
+                self.log.with_object(n).debug(
+                    'Metrics for node: %s', new_metrics)
+                metrics[n.uuid] = new_metrics
+
             except exceptions.ReadException:
                 self.log.with_object(n).warning(
                     'Refreshing metrics for node failed')
@@ -77,22 +79,35 @@ class Scheduler(object):
         self.metrics = metrics
         self.metrics_updated = time.time()
 
-    def _has_sufficient_cpu(self, cpus, node):
-        max_cpu = (self.metrics[node].get('cpu_max', 0) *
-                   config.get('CPU_OVERCOMMIT_RATIO'))
+    def _has_sufficient_cpu(self, log_ctx, cpus, node):
+        hard_max_cpus = (self.metrics[node].get(
+            'cpu_max', 0) * config.CPU_OVERCOMMIT_RATIO)
         current_cpu = self.metrics[node].get('cpu_total_instance_vcpus', 0)
-        if current_cpu + cpus > max_cpu:
+
+        if current_cpu + cpus > hard_max_cpus:
+            log_ctx.with_fields({
+                'node': node,
+                'current_cpus': current_cpu,
+                'requested_cpus': cpus,
+                'hard_max_cpus': hard_max_cpus
+            }).debug('Scheduling on node would exceed hard maximum CPUs')
             return False
+
         return True
 
-    def _has_sufficient_ram(self, memory, node):
+    def _has_sufficient_ram(self, log_ctx, memory, node):
         # There are two things to track here... We must always have
         # RAM_SYSTEM_RESERVATION gb of RAM for operating system tasks -- assume
         # there is no overlap with existing VMs when checking this. Note as
         # well that metrics are in MB...
         available = (self.metrics[node].get('memory_available', 0) -
-                     (config.get('RAM_SYSTEM_RESERVATION') * 1024))
+                     (config.RAM_SYSTEM_RESERVATION * 1024))
         if available - memory < 0.0:
+            log_ctx.with_fields({
+                'node': node,
+                'available': available,
+                'requested_memory': memory
+            }).debug('Insufficient memory')
             return False
 
         # ...Secondly, if we're using KSM and over committing memory, we
@@ -100,21 +115,35 @@ class Scheduler(object):
         instance_memory = (
             self.metrics[node].get('memory_total_instance_actual', 0) + memory)
         if (instance_memory / self.metrics[node].get('memory_max', 0) >
-                config.get('RAM_OVERCOMMIT_RATIO')):
+                config.RAM_OVERCOMMIT_RATIO):
+            log_ctx.with_fields({
+                'node': node,
+                'instance_memory': instance_memory,
+                'memory_max': self.metrics[node].get('memory_max', 0),
+                'overcommit_ratio': config.RAM_OVERCOMMIT_RATIO
+            }).debug('KSM overcommit ratio exceeded')
             return False
 
         return True
 
-    def _has_sufficient_disk(self, instance, node):
+    def _has_sufficient_disk(self, log_ctx, inst, node):
         requested_disk = 0
-        for disk in instance.disk_spec:
+        for disk in inst.disk_spec:
             # TODO(mikal): this ignores "sizeless disks", that is ones that
             # are exactly the size of their base image, for example CD ROMs.
             if 'size' in disk:
                 if not disk['size'] is None:
                     requested_disk += int(disk['size'])
 
-        if requested_disk > (int(self.metrics[node].get('disk_free', '0')) / 1024 / 1024 / 1024):
+        disk_free = int(self.metrics[node].get(
+            'disk_free_instances', '0')) / GiB
+        disk_free -= config.MINIMUM_FREE_DISK
+        if requested_disk > disk_free:
+            log_ctx.with_fields({
+                'node': node,
+                'requested_disk_gb': requested_disk,
+                'disk_free_gb': disk_free,
+            }).debug('Node has insufficient disk')
             return False
         return True
 
@@ -124,11 +153,10 @@ class Scheduler(object):
 
         # Find number of matching networks on each node. We need to be careful
         # how we do this to avoid repeatedly scanning the etcd repository.
-        per_node = {}
-        for inst in virt.Instances([]):
+        per_node = defaultdict(list)
+        for inst in instance.Instances([]):
             n = inst.placement
             if n.get('node'):
-                per_node.setdefault(n['node'], [])
                 per_node[n['node']].append(inst)
 
         candidates_network_matches = {}
@@ -138,8 +166,14 @@ class Scheduler(object):
             # Make a list of networks for the node
             present_networks = []
             for inst in per_node.get(n, []):
-                for ni in networkinterface.interfaces_for_instance(inst):
-                    if ni.network_uuid not in present_networks:
+                for iface_uuid in inst.interfaces:
+                    ni = networkinterface.NetworkInterface.from_db(iface_uuid)
+                    if not ni:
+                        LOG.with_fields({
+                            'instance': inst.uuid,
+                            'networkinterface': iface_uuid
+                        }).error('Interface missing while attempting schedule')
+                    elif ni.network_uuid not in present_networks:
                         present_networks.append(ni.network_uuid)
 
             # Count the requested networks present on this node
@@ -148,10 +182,10 @@ class Scheduler(object):
                     candidates_network_matches[n] += 1
 
         # Store candidate nodes keyed by number of matches
-        candidates_by_network_matches = {}
+        candidates_by_network_matches = defaultdict(list)
         for n in candidates:
             matches = candidates_network_matches[n]
-            candidates_by_network_matches.setdefault(matches, []).append(n)
+            candidates_by_network_matches[matches].append(n)
 
         # Find maximum matches of networks on a node
         max_matches = max(candidates_by_network_matches.keys())
@@ -167,35 +201,9 @@ class Scheduler(object):
         # Return list of candidates that has maximum networks
         return candidates_by_network_matches[max_matches]
 
-    def _find_most_matching_images(self, requested_images, candidates):
-        # Determine number of matching images per node
-        candidates_image_matches = {}
-        for n in candidates:
-            candidates_image_matches[n] = 0
-
-        for image in requested_images:
-            for i in images.Images(filters=[partial(images.url_filter, image),
-                                            baseobject.active_states_filter]):
-                candidates_image_matches[i.node] += 1
-
-        # Create dict of candidate lists keyed by number of image matches
-        candidates_by_image_matches = {}
-        for n in candidates:
-            matches = candidates_image_matches[n]
-            candidates_by_image_matches.setdefault(matches, [])
-            candidates_by_image_matches[matches].append(n)
-
-        # If no matches, return the original candidate list
-        if len(candidates_by_image_matches) == 0:
-            return candidates
-
-        # Return all candidates that have the highest number of image matches
-        max_matches = max(candidates_by_image_matches.keys())
-        return candidates_by_image_matches[max_matches]
-
-    def place_instance(self, instance, network, candidates=None):
-        with util.RecordedOperation('schedule', instance):
-            log_ctx = self.log.with_object(instance)
+    def place_instance(self, inst, network, candidates=None):
+        with util_general.RecordedOperation('schedule', inst):
+            log_ctx = self.log.with_object(inst)
 
             # Refresh metrics if its too old, or there are no nodes.
             diff = time.time() - self.metrics_updated
@@ -210,8 +218,8 @@ class Scheduler(object):
             if candidates:
                 log_ctx.with_field('candidates', candidates).info(
                     'Scheduling: forced candidates')
-                instance.add_event('schedule',
-                                   'Forced candidates', None, str(candidates))
+                inst.add_event('schedule',
+                               'Forced candidates', None, str(candidates))
                 for n in candidates:
                     if n not in self.metrics:
                         raise exceptions.CandidateNodeNotFoundException(n)
@@ -221,100 +229,106 @@ class Scheduler(object):
                     candidates.append(n)
             log_ctx.with_field('candidates', candidates).info(
                 'Scheduling: Initial candidates')
-            instance.add_event('schedule',
-                               'Initial candidates', None, str(candidates))
+
+            # Ensure all specified nodes are hypervisors
+            for n in list(candidates):
+                inst.add_event(
+                    'schedule',
+                    'Is hypervisor: %s' % self.metrics[n].get(
+                        'is_hypervisor', False),
+                    None, str(n))
+                if not self.metrics[n].get('is_hypervisor', False):
+                    candidates.remove(n)
+            log_ctx.with_field('candidates', candidates).info(
+                'Scheduling: Are hypervisors')
+
+            inst.add_event('schedule',
+                           'Initial candidates', None, str(candidates))
             if not candidates:
                 raise exceptions.LowResourceException('No nodes with metrics')
 
             # Can we host that many vCPUs?
-            for n in copy.copy(candidates):
+            for n in list(candidates):
                 max_cpu = self.metrics[n].get('cpu_max_per_instance', 0)
-                if instance.cpus > max_cpu:
+                if inst.cpus > max_cpu:
                     candidates.remove(n)
             log_ctx.with_field('candidates', candidates).info(
                 'Scheduling: have enough actual CPU')
-            instance.add_event('schedule',
-                               'Have enough actual CPU', None, str(candidates))
+            inst.add_event('schedule',
+                           'Have enough actual CPU', None, str(candidates))
             if not candidates:
                 raise exceptions.LowResourceException(
                     'Requested vCPUs exceeds vCPU limit')
 
             # Do we have enough idle CPU?
-            for n in copy.copy(candidates):
-                if not self._has_sufficient_cpu(
-                        instance.cpus, n):
+            for n in list(candidates):
+                if not self._has_sufficient_cpu(log_ctx, inst.cpus, n):
                     candidates.remove(n)
             log_ctx.with_field('candidates', candidates).info(
                 'Scheduling: have enough idle CPU')
-            instance.add_event('schedule',
-                               'Have enough idle CPU', None, str(candidates))
+            inst.add_event('schedule',
+                           'Have enough idle CPU', None, str(candidates))
             if not candidates:
                 raise exceptions.LowResourceException(
                     'No nodes with enough idle CPU')
 
             # Do we have enough idle RAM?
-            for n in copy.copy(candidates):
-                if not self._has_sufficient_ram(
-                        instance.memory, n):
+            for n in list(candidates):
+                if not self._has_sufficient_ram(log_ctx, inst.memory, n):
                     candidates.remove(n)
             log_ctx.with_field('candidates', candidates).info(
                 'Scheduling: Have enough idle RAM')
-            instance.add_event('schedule',
-                               'Have enough idle RAM', None, str(candidates))
+            inst.add_event('schedule',
+                           'Have enough idle RAM', None, str(candidates))
             if not candidates:
                 raise exceptions.LowResourceException(
                     'No nodes with enough idle RAM')
 
             # Do we have enough idle disk?
-            for n in copy.copy(candidates):
-                if not self._has_sufficient_disk(instance, n):
+            for n in list(candidates):
+                if not self._has_sufficient_disk(log_ctx, inst, n):
                     candidates.remove(n)
             log_ctx.with_field('candidates', candidates).info(
                 'Scheduling: Have enough idle disk')
-            instance.add_event('schedule',
-                               'Have enough idle disk', None, str(candidates))
+            inst.add_event('schedule',
+                           'Have enough idle disk', None, str(candidates))
             if not candidates:
                 raise exceptions.LowResourceException(
                     'No nodes with enough disk space')
 
-            # What nodes have the highest number of networks already present?
-            if network:
-                requested_networks = []
-                for net in network:
-                    network_uuid = net['network_uuid']
-                    if network_uuid not in requested_networks:
-                        requested_networks.append(network_uuid)
+            # Calc pseudo CPU load from inst affinity
+            pseudo_load = defaultdict(int)
+            cpu_affinity = inst.affinity.get('cpu')
+            if cpu_affinity:
+                for i in instance.Instances([instance.healthy_states_filter]):
+                    if i.uuid == inst.uuid or not i.tags:
+                        continue
+                    for tag, val in cpu_affinity.items():
+                        if tag in i.tags:
+                            # Allow for unplaced healthy instances
+                            n = i.placement.get('node')
+                            if n:
+                                pseudo_load[n] += int(val)
 
-                candidates = self._find_most_matching_networks(
-                    requested_networks, candidates)
-                log_ctx.with_field('candidates', candidates).info(
-                    'Scheduling: Have most matching networks')
-                instance.add_event('schedule', 'Have most matching networks',
-                                   None, str(candidates))
+                inst.add_event('schedule', 'CPU load with affinity',
+                               None, str(dict(pseudo_load)))
 
-            # What nodes have the base image already?
-            requested_images = []
-            for disk in instance.disk_spec:
-                if disk.get('base'):
-                    img = images.Image.new(disk['base'])
-                    requested_images = img.url
+            # Order candidates by current CPU load
+            by_load = defaultdict(list)
+            for n in list(candidates):
+                load = math.floor(self.metrics[n].get('cpu_load_1', 0))
+                load -= pseudo_load.get(n, 0)
+                by_load[load].append(n)
+            log_ctx.with_field('by_load', by_load).info(
+                'Scheduling: Adjusted CPU load')
 
-            candidates = self._find_most_matching_images(
-                requested_images, candidates)
-            log_ctx.with_field('candidates', candidates).info(
-                'Scheduling: Have most matching images')
-            instance.add_event('schedule', 'Have most matching images',
-                               None, str(candidates))
-
-            # Avoid allocating to network node if possible
-            net_node = get_network_node()
-            if len(candidates) > 1 and net_node.uuid in candidates:
-                candidates.remove(net_node.uuid)
-                log_ctx.with_field('candidates', candidates).info(
-                    'Scheduling: Are non-network nodes')
-                instance.add_event('schedule', 'Are non-network nodes',
-                                   None, str(candidates))
+            lowest_load = sorted(by_load)[0]
+            candidates = by_load[lowest_load]
+            inst.add_event('schedule', 'Have lowest CPU load',
+                           None, str(candidates))
 
             # Return a shuffled list of options
             random.shuffle(candidates)
+            inst.add_event('schedule', 'Final candidates',
+                           None, str(candidates))
             return candidates
